from .constants import PRIMITIVE_TYPES
from argon_88_sk.utils.templates import XML_ITEM


class Formatter:

    @staticmethod
    def move_line(string, indent):
        """

        :param string:
        :param indent:
        :return:
        """
        return "\t" * indent + string

    def to_json(self, obj, dumps):
        """

        :param obj:
        :param dumps:
        :return:
        """
        items_repr = ""

        for k, v in obj.items():
            if type(v) in PRIMITIVE_TYPES:
                items_repr += f"\t{dumps(k)}: {dumps(v)},\n"
                continue

            items_repr += f"\t{dumps(k)}: {{\n"

            for line in dumps(v).split("\n")[1:]:
                items_repr += f"{self.move_line(line, 1)}\n"

        return items_repr

    def to_xml(self, obj, dumps):
        items_repr = ""

        for k, v in obj.items():
            items_repr += self.move_line(
                XML_ITEM.format(
                    key=dumps(k),
                    value=dumps(v)
                ), 1)

        return items_repr


import re
from abc import ABC, abstractmethod
from types import NoneType, FunctionType, MethodType, CodeType, ModuleType,\
    CellType, BuiltinMethodType, BuiltinFunctionType, WrapperDescriptorType, \
    MethodDescriptorType, MappingProxyType, GetSetDescriptorType, MemberDescriptorType, GeneratorType
from typing import Any, IO, Hashable, Collection, Iterable

from argon_88_sk.utils.constants import TYPE_MAPPING
from argon_88_sk.utils.helpers import Formatter


class Serializer(ABC):
    _IGNORED_FIELDS: set[str] = (
        '__weakref__',
        '__subclasshook__',
        '__dict__',
        '__doc__'
    )
    _IGNORED_FIELD_TYPES: set[str] = (
        BuiltinFunctionType, BuiltinMethodType,
        WrapperDescriptorType, MethodDescriptorType,
        MappingProxyType, GetSetDescriptorType,
        MemberDescriptorType
    )
    formatter = Formatter()

    @staticmethod
    def _get_key(value: Hashable, obj: dict):
        return [key for key in obj if obj[key] == value][0]

    @classmethod
    def _type_from_str(cls, s: str, pattern: str) -> type:
        if not re.search(pattern, s):
            return NoneType

        return TYPE_MAPPING[re.search(pattern, s).group(1)]

    def get_items(self, obj) -> dict[str, Any]:
        """

        :param obj:
        :return:
        """
        if isinstance(obj, (BuiltinFunctionType, BuiltinMethodType)):
            return {}

        if isinstance(obj, dict):
            return obj

        elif isinstance(obj, Collection):
            return dict(enumerate(obj))

        elif isinstance(obj, CodeType):
            return {
                "argcount": obj.co_argcount,
                "posonlyargcount": obj.co_posonlyargcount,
                "kwonlyargcount": obj.co_kwonlyargcount,
                "nlocals": obj.co_nlocals,
                "stacksize": obj.co_stacksize,
                "flags": obj.co_flags,
                "code": obj.co_code,
                "consts": obj.co_consts,
                "names": obj.co_names,
                "varnames": obj.co_varnames,
                "filename": obj.co_filename,
                "name": obj.co_name,
                "firstlineno": obj.co_firstlineno,
                "lnotab": obj.co_lnotab,
                "freevars": obj.co_freevars,
                "cellvars": obj.co_cellvars,
            }

        elif isinstance(obj, property):
            return {
                "getter": obj.fget,
                "setter": obj.fset,
                "deleter": obj.fdel
            }

        elif isinstance(obj, GeneratorType):
            a = []
            for i in obj:
                a.append(i)
            return{
                "values":a
            }

        elif isinstance(obj, FunctionType):
            if obj.__closure__ and "__class__" in obj.__code__.co_freevars:
                closure = ([... for _ in obj.__closure__])
            elif obj.__closure__:
                closure = ([cell.cell_contents for cell in obj.__closure__])
            else:
                closure = None

            return {
                "argcount": obj.__code__.co_argcount,
                "posonlyargcount": obj.__code__.co_posonlyargcount,
                "kwonlyargcount": obj.__code__.co_kwonlyargcount,
                "nlocals": obj.__code__.co_nlocals,
                "stacksize": obj.__code__.co_stacksize,
                "flags": obj.__code__.co_flags,
                "code": obj.__code__.co_code,
                "consts": obj.__code__.co_consts,
                "names": obj.__code__.co_names,
                "varnames": obj.__code__.co_varnames,
                "filename": obj.__code__.co_filename,
                "name": obj.__code__.co_name,
                "firstlineno": obj.__code__.co_firstlineno,
                "lnotab": obj.__code__.co_lnotab,
                "freevars": obj.__code__.co_freevars,
                "cellvars": obj.__code__.co_cellvars,
                "globals": {
                    k: obj.__globals__[k]
                    for k in (
                        set(
                            k for k, v in obj.__globals__.items()
                            if isinstance(v, ModuleType)
                        ) |
                        set(obj.__globals__) &
                        set(obj.__code__.co_names) -
                        {obj.__name__}
                    )
                },
                "closure": closure,
                "qualname": obj.__qualname__
            }

        elif isinstance(obj, MethodType):
            return {
                "__func__": obj.__func__,
                "__self__": obj.__self__
            }

        elif issubclass(type(obj), type):
            return {
                'name': obj.__name__,
                'mro': tuple(obj.mro()[1:-1]),
                'attrs': {
                    k: v for k, v in obj.__dict__.items()
                    if (
                        k not in self._IGNORED_FIELDS and
                        type(v) not in self._IGNORED_FIELD_TYPES
                    )
                }
            }

        elif issubclass(type(obj), ModuleType):
            return {'name': obj.__name__}

        elif isinstance(obj, staticmethod):
            return self.get_items(obj.__func__)

        elif isinstance(obj, classmethod):
            return self.get_items(obj.__func__)

        else:
            return {
                'class': obj.__class__,
                'attrs': {
                    k: v for k, v in obj.__dict__.items()
                    if (
                        k not in self._IGNORED_FIELDS and
                        type(k) not in self._IGNORED_FIELD_TYPES
                    )
                }
            }

    # noinspection PyArgumentList,PyUnresolvedReferences
    def create_object(self, obj_type: type, obj_data):
        if issubclass(obj_type, dict):
            return obj_data

        elif issubclass(obj_type, GeneratorType):
            return (x for x in obj_data["values"])

        elif issubclass(obj_type, Iterable):
            return obj_type(obj_data.values())

        elif issubclass(obj_type, CodeType):
            return CodeType(*list(obj_data.values()))

        elif issubclass(obj_type, FunctionType):
            if obj_data.get('closure'):
                closure = tuple([CellType(x) for x in obj_data.get('closure')])
            elif obj_data.get('closure') and '__class__' in obj_data.get('freevars'):
                closure = tuple([CellType(...) for _ in obj_data.get('closure')])
            else:
                closure = tuple()

            obj = FunctionType(
                code=CodeType(*list(obj_data.values())[:16]),
                globals=obj_data.get('globals'),
                name=obj_data['name'],
                closure=closure
            )
            obj.__qualname__ = obj_data.get('qualname')
            obj.__globals__[obj.__name__] = obj

            return obj

        elif issubclass(obj_type, property):
            return property(fget=obj_data.get("getter"), fdel=obj_data.get("deleter"), fset=obj_data.get("setter"))

        elif issubclass(obj_type, MethodType):
            return MethodType(
                obj_data.get('__func__'),
                obj_data.get('__self__'),
            )

        elif issubclass(obj_type, (staticmethod, classmethod)):
            return self.create_object(FunctionType, obj_data)

        elif issubclass(obj_type, type):
            obj = type(obj_data.get('name'), obj_data.get('mro'), obj_data.get('attrs'))

            try:
                obj.__init__.__closure__[0].cell_contents = obj
            except (AttributeError, IndexError):
                ...

            return obj

        elif issubclass(obj_type, ModuleType):
            return __import__(obj_data.get('name'))

        else:
            obj = object.__new__(obj_data.get('class'))
            obj.__dict__ = obj_data.get('attrs')

            return obj

    def dump(self, obj: Any, fp: IO[str]) -> None:
        """Dumps an object to .json file.

        :param obj: object to dump.
        :param fp: json file object.
        """
        fp.write(self.dumps(obj))

    def load(self, fp: IO[str]):
        """Loads an object from .json file.

        :param fp: json file object to extract object from.
        :return: deserialized Python object.
        """
        return self.loads(fp.read())

    @abstractmethod
    def dumps(self, obj) -> str:
        """Dumps an object to a string and returns the string.

        :param obj: object to dump.
        :return: string containing serialized (dumped) object.
        """
        raise NotImplementedError

    @abstractmethod
    def loads(self, s):
        """Loads an object from a string and returns it.

        :param s: string to extract object from.
        :return: deserialized Python object.
        """
        raise NotImplementedError


import argparse

from argon_88_sk.base import Serializer
from argon_88_sk.serialize import JSONSerializer, XMLSerializer


def main():
    parser = argparse.ArgumentParser(prog="argon_88_sk")
    parser.add_argument('file_from')
    parser.add_argument('file_to')
    parser.add_argument('format_from')
    parser.add_argument('format_to')

    args = parser.parse_args()
    file_from, file_to, format_from, format_to = (
        args.file_from,
        args.file_to,
        args.format_from,
        args.format_to
    )

    format_mapping: dict[str, Serializer] = {
        'json': JSONSerializer(),
        'xml': XMLSerializer()
    }

    with open(file_from, 'r') as ff, open(file_to, 'w+') as ft:
        ser_from: Serializer = format_mapping[format_from]
        ser_to: Serializer = format_mapping[format_to]

        ser_to.dump(ser_from.load(ff), ft)


if __name__ == '__main__':
    filename = "/home/artem/Desktop/file1.json"

    with open(filename, 'w+') as fp:
        JSONSerializer().dump({
            1: "123",
            None: [True, False, "sdf", -11],
            90: 1,
        }, fp)

    main()



from types import (
    NoneType, FunctionType, LambdaType,
    MethodType, CodeType, CellType, ModuleType, EllipsisType, GeneratorType
)

PRIMITIVE_TYPES: tuple = (int, float, complex, str, bool, NoneType, EllipsisType)
TYPE_MAPPING = {
    'int': int,
    'float': float,
    'complex': complex,
    'str': str,
    'bool': bool,
    'NoneType': NoneType,
    'ellipsis': EllipsisType,
    'bytes': bytes,
    'list': list,
    'tuple': tuple,
    'set': set,
    'dict': dict,
    'code': CodeType,
    'cell': CellType,
    'function': FunctionType,
    'lambda': LambdaType,
    'method': MethodType,
    'staticmethod': staticmethod,
    'classmethod': classmethod,
    'type': type,
    'module': ModuleType,
    'object': object,
    'property': property,
    "generator": GeneratorType
}


"""Serialization of Lab_3"""
import re
from types import NoneType, EllipsisType
from typing import Iterator

from .base import Serializer
from argon_88_sk.utils.templates import JSON, XML, XML_PRIMITIVE
from argon_88_sk.utils.constants import PRIMITIVE_TYPES, TYPE_MAPPING


class JSONSerializer(Serializer):
    """JSON serializer class."""
    _TYPE_PATTERN: str = r"<class '(\w\S+)'>_"
    _KEYWORDS: dict[None | bool, str] = {
        None: 'null',
        True: 'true',
        False: 'false'
    }

    def __init__(self):
        super().__init__()

    @classmethod
    def _to_number(cls, s: str) -> int | float | complex | None:
        for num_type in (int, float, complex):
            try:
                return num_type(s)
            except (ValueError, TypeError):
                pass

    def _load_from_json(self, template: str) -> dict:
        """Takes a string of specific format (visit ``utils.templates``
        for more clarity) as an input, loads object data,
        and returns it in the form of dict.

        :param template: string template to retrieve object from.
        :return: dictionary with object data.
        """
        obj: dict = {}
        lines: list[str] = template.split("\n")
        it: Iterator[str] = enumerate(lines)

        for i, line in it:
            if not re.search(r'\s*(.+):\s*([^,]*)', line):
                continue

            key, value = re.search(r'\s*(.+):\s*([^,]*)', line).groups()

            if value != "{":
                obj[self.loads(key)] = self.loads(value)

            elif value == "{" and "<class" not in key:
                brackets = 1
                start = i + 1

                while brackets and i < len(lines) - 1:
                    i, line = next(it, None)
                    brackets += ("{" in lines[i]) - ("}" in lines[i])

                obj[self.loads(key)] = self.loads('\n'.join(lines[start:i]))

        return obj

    def dumps(self, obj) -> str:
        """Dumps an object to a string and returns the string.
        Dumping is done via general JSON object template. It can
        overcomplicate simple structure serialization, but can be
        applied to much larger scale of python objects.

        :param obj: object to dump.
        :return: string containing serialized (dumped) object.
        """
        if type(obj) == str:
            return f'"{obj}"'
        if type(obj) == type(Ellipsis):
            return ' '
        if type(obj) in (int, float, complex):
            return str(obj)
        if type(obj) in [bool, NoneType]:
            return self._KEYWORDS[obj]

        return JSON.format(
            type=type(obj) if type(obj) in TYPE_MAPPING.values() else object,
            id=id(obj),
            items=self.formatter.to_json(self.get_items(obj), self.dumps)
        )

    def loads(self, s: str):
        """Loads an object from a string and returns it.
        Operates using JSON template from ``utils.templates``.
        However, primitive types are serialized without this
        or any other template.

        :param s: string to extract object from.
        :return: deserialized Python object.
        """
        if not len(s):
            return

        if s == ' ':
            return ...
        if s.startswith('"'):
            return s.strip('"')
        if s in self._KEYWORDS.values():
            return self._get_key(s, self._KEYWORDS)
        if self._to_number(s) is not None:
            return self._to_number(s)

        return self.create_object(
            self._type_from_str(s, self._TYPE_PATTERN),
            self._load_from_json(s)
        )


class XMLSerializer(Serializer):
    _TYPE_PATTERN: str = r'type="(\w+)"'

    def _get_tag(self, tagname: str, lines) -> str:
        counter = 1
        it = enumerate(lines)

        for i, line in it:
            if not counter:
                return lines[:i]

            counter += bool(re.search(rf"<{tagname}.*>", line.strip("\t\n ")))
            counter -= bool(re.search(rf"</{tagname}>", line.strip("\t\n ")))

    def _load_from_xml(self, template: str) -> dict:
        obj: dict = {}
        lines: list[str] = template.split("\n")
        it: Iterator[str] = enumerate(lines)

        for i, line in it:
            if "<item>" == line.strip("\t\n "):
                item = self._get_tag("item", lines[i+1:])
                key = self._get_tag("key", item[1:])
                value = self._get_tag("value", item[len(key)+2:])

                obj[self.loads("\n".join(key[:-1]))] = self.loads("\n".join(value[:-1]))

                [next(it, None) for _ in range(len(item))]

        return obj

    def dumps(self, obj) -> str:
        """Dumps an object to a string and returns the string.

        Dumping is done via string templates with XML prefix in
        ``utils.templates`` module.

        :param obj: object to dump.
        :return: string containing serialized (dumped) object.
        """
        if type(obj) in PRIMITIVE_TYPES:
            obj_type = self._get_key(type(obj), TYPE_MAPPING)
            return f'<primitive type="{obj_type}">{obj}</primitive>'

        return XML.format(
            type=self._get_key(type(obj), TYPE_MAPPING) if type(obj) in TYPE_MAPPING.values() else "object",
            id=id(obj),
            items=self.formatter.to_xml(self.get_items(obj), self.dumps)
        )

    def loads(self, s):
        """Loads an object from a string and returns it.

        Operates using templates with XML prefix from ``utils.templates``.

        :param s: string to extract object from.
        :return: deserialized Python object.
        """
        if not len(s):
            return

        if "primitive" in s.split("\n")[0]:
            obj_data = re.search(
                XML_PRIMITIVE.format(
                    type="\w+",
                    obj="(.*)"
                ), s).group(1)
            obj_type = self._type_from_str(
                s=s.split("\n")[0],
                pattern=self._TYPE_PATTERN
            )

            if obj_type == NoneType:
                return None

            if obj_type == bool:
                return obj_data == "True"

            if obj_type == EllipsisType:
                return ...

            return obj_type(obj_data)

        return self.create_object(
            self._type_from_str(s, self._TYPE_PATTERN),
            self._load_from_xml(s)
        )


from setuptools import find_packages, setup

setup(
    name="argon_88_sk",
    version="0.1.2",
    packages=find_packages(include=["argon_88_sk", "argon_88_sk.*"]),
    description="XML & JSON serialization tools at BSUIR 2023 spring semester.""",
    author="Me",
    license="MIT",
    entry_points={
        'console_scripts': ['argon_88_sk=argon_88_sk.lib:main']
    }
)


JSON = ('''{{
    "{type}_{id:x}": {{
    {items}
    }}
}}''')
XML = ('''
<object type="{type}" id="{id:x}">
{items}
</object>
''')
XML_PRIMITIVE = '<primitive type="{type}">{obj}</primitive>'
XML_ITEM = ('''
<item>
    <key>
        {key}
    </key>
    <value>
        {value}
    </value>
</item>
''')